/*******************************************************************************
 *	BdsImportData.cc	Simple Data import from old Blacknest Archive
 *			T.Barnaby,	BEAM Ltd,	2008-06-23
 *******************************************************************************
 */
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <errno.h>
#include <BdsD.h>
#include <BdsC.h>
#include <BMysql.h>
#include <BFile.h>
#include <BDir.h>
#include <BdsLib.h>
#include <BdsDataLib.h>
#include <BTable.h>
#include <BdsDataFileSeed.h>

using namespace std;
using namespace Bds;


// An import file
class ImportFile {
public:
			ImportFile(BString fileName = "") : fileName(fileName), file(0){ }
			~ImportFile(){ close(); }

	void		close(){ if(file) delete file; file = 0; }	

	BString		fileName;
	DataInfo	dataInfo;
	DataFile*	file;
};

// An import channel
class ImportChannel {
public:
			ImportChannel(ImportFile& file, BUInt fileChannel) : file(file), fileChannel(fileChannel), complete(0){}
			ImportChannel(const ImportChannel& c) : file(c.file), fileChannel(c.fileChannel), complete(c.complete){}
	
	ImportChannel&	operator=(const ImportChannel& c){ file = c.file; fileChannel = c.fileChannel; complete = c.complete; return *this; }
	ImportFile&	file;
	BUInt		fileChannel;
	int		complete;
};

class Config {
public:
	BString			hostName;			///< The Hostname
	BString			network;			///< The network
	BString			array;				///< The Array
	BString			source;				///< The source
	BString			format;				///< The import data file format
	BString			description;			///< A description to add
	BStringList		fileNames;			///< The list of file names to import
	BStringList		channels;			///< An optional list of channels names to import, overides data file provided names
	int			synchronous;			///< All data is syncronously sampled
	int			validate;			///< Validate the data files
	int			validateMissingBlocks;		///< Validate any missing blocks
	int			validateTimeBackwards;		///< Validate if time goes nackwards
	int			validateFilenameTime;		///< Validate file name time
	int			validateMetaData;		///< Validate that MetaData exists
	int			validateCorruptions;		///< Validate any curruptions
	int			reorder;			///< Re-order data in files
	int			deleteDuplicates;		///< Delete any duplicate blocks
	int			warnings;			///< Print out any warnings
	int			verbose;			///< Provide diagnostic information
	int			dryRun;				///< Validate files, but do not import
	int			formatList;			///< List the data formats supported
	BStringList		warningList;			///< List of warnings to add
	int			sequential;			///< Sequential in time files
	BUInt			numBlocks;			///< Max number of blocks to import (for testing)
};

void printError(BString str){
	cerr << "Error, " << str << "\n";
}

void printWarning(BString str){
	cerr << "Warning, " << str << "\n";
}


BError validateMetaData(DataAddAccess& bds, DataInfo& dataInfo){
	BError			err;
	ChannelInfos		channelInfos;

	if(err = bds.dataGetChannelInfo(dataInfo, channelInfos)){
		err.set(ErrorValidateMetaData, BString("Incomplete meta data for channels available: ") + err.getString());
		return err;
	}
	
	if(channelInfos.channels.number() != dataInfo.channels.size()){
		if(channelInfos.channels.number())
			err.set(ErrorValidateMetaData, "Incomplete meta data for channels available");
		else
			err.set(ErrorValidateMetaData, "No meta data for channels available");
	}

	return err;
}

// Import a set of data files
BError importData(DataAddAccess& bds, Config& config){
	BError				err;
	BError				e;
	BList<ImportFile>		importFiles;			// The list of import files
	BArray<BArray<ImportChannel> >	importChannels;			// Import channels by segment,channel
	DataInfo			dataInfo;
	DataHandle			dataHandle;
	DataBlock			dataBlock;
	BList<BError>			errors;
	BUInt32				blockNumber = 0;
	BIter				i;
	BIter				di;
	BUInt				chan;
	BUInt				seg;
	BUInt				chanFile;
	BString				format;
	DataFile::Options		options = 0;
	double				sampleRate = 0.0;
	int				fileNumber = 0;
	int				warningNumber = 0;
	BString				errString;
	int				userDefinedChannels = 0;
	BTimeStamp			lastTime;

	// Setup data file options	
	if(config.validate)
		options |= DataFile::OptionValidate;
	if(config.validateCorruptions)
		options |= DataFile::OptionValidateCorruptions;
	if(config.validateFilenameTime)
		options |= DataFile::OptionFileNameProcess;
	if(config.reorder)
		options |= DataFile::OptionReorder;
	if(config.deleteDuplicates)
		options |= DataFile::OptionDeleteDuplicates;
	
	// Setup list of import files
	for(config.fileNames.start(i); !config.fileNames.isEnd(i); config.fileNames.next(i)){
		importFiles.append(ImportFile(config.fileNames[i]));
	}

	// Setup DataInfo
	dataInfo.array = config.array;
	dataInfo.description = config.description;
	dataInfo.synchronous = config.synchronous;

	for(config.channels.start(i); !config.channels.isEnd(i); config.channels.next(i)){
		DataChannel		dc;
		BArray<DataChannel>	dcList;

		dc.network = config.network;
		dc.station = config.channels[i].pullToken(":");
		dc.channel = config.channels[i];
		dc.source = config.source;

		dcList.append(dc);
		dataInfo.channels.append(dcList);
		userDefinedChannels = 1;
	}

	for(config.warningList.start(i); !config.warningList.isEnd(i); config.warningList.next(i)){
		dataInfo.info[BString("importWarning") + warningNumber++] = config.warningList[i];
	}
	
	if(importFiles.size() > 1)
		format = "API-CM";
	else
		format = "API-SM";

	// Go through all files finding Start and end times of all channels and validate all files
	for(importFiles.start(i), chan = 0, seg = 0; !importFiles.isEnd(i); importFiles.next(i)){
		ImportFile&	f = importFiles[i];

		importChannels.resize(seg + 1);

//		printf("DoFile: %d: %s\n", fileNumber, f.fileName.retStr());	
		dataInfo.info[BString("importFile") + fileNumber + ".fileName"] = f.fileName;
		
		if(err = dataFormats.formatGet(config.format, f.file))
			return err.set(1, BString("Unsupported file format: ") + config.format);

		// Force channel multiplexed if one of the data files is channel multiplexed
		if(f.file->getDataOrder() == DataFile::DataOrderChannel){
			format = "API-CM";
		}

		if(config.verbose){
			std::cout << "Process File: "  << f.fileName <<	" Chan: " << chan << " Seg: " << seg << "\n";
		}

		if(err = f.file->open(f.fileName, "r"))
			return err;

		errors.clear();
		if(err = f.file->getInfo(f.dataInfo, options, errors)){
			if(err.getErrorNo() == ErrorValidate){
				err.set(0, "");
				for(errors.start(di); !errors.isEnd(di); errors.next(di)){
					switch(errors[di].getErrorNo()){
					case ErrorValidateFilenameTime:
						errString = BString("Filename Time, ") + f.fileName + ", " + errors[di].getErrorNo() + ", " + errors[di].getString();
						if(config.validateFilenameTime){
							err.set(ErrorValidate, "Validation error");
							printError(errString);
						}
						else if(config.warnings){
							printWarning(errString);
						}
						break;
					case ErrorValidateMissingBlocks:
						errString = BString("Missing Blocks, ") + f.fileName + ", " + errors[di].getErrorNo() + ", " + errors[di].getString();
						if(config.validateMissingBlocks){
							err.set(ErrorValidate, "Validation error");
							printError(errString);
						}
						else if(config.warnings){
							printWarning(errString);
						}
						break;
					case ErrorValidateTimeBackwards:
						errString = BString("Time Backwards, ") + f.fileName + ", " + errors[di].getErrorNo() + ", " + errors[di].getString();
						if(config.validateTimeBackwards){
							err.set(ErrorValidate, "Validation error");
							printError(errString);
						}
						else if(config.warnings){
							printWarning(errString);
						}
						break;
					case ErrorValidateFix:
						errString = BString("Curruption, ") + f.fileName + ", " + errors[di].getErrorNo() + ", " + errors[di].getString();
						if(config.validateCorruptions){
							err.set(ErrorValidate, "Validation error");
							printError(errString);
						}
						else if(config.warnings){
							printWarning(errString);
						}
						break;
					case ErrorValidateReorder:
						if(config.warnings){
							errString = BString("Reordered, ") + f.fileName + ", " + errors[di].getErrorNo() + ", " + errors[di].getString();
							printWarning(errString);
						}
						break;
					default:
						errString = BString("Misc, ") + f.fileName + ", " + errors[di].getErrorNo() + ", " + errors[di].getString();
						err.set(ErrorValidate, "Validation error");
						printError(errString);
						break;
					}
					
					dataInfo.info[BString("importWarning") + warningNumber++] = errString;
				}
			}
			
			if(err)
				return err;
		}

		if(config.verbose){
			std::cout << "\tStartTime: " << f.dataInfo.startTime.getString() << " EndTime: " << f.dataInfo.endTime.getString() << "\n";
		}

		if(!dataInfo.startTime.isSet() || (dataInfo.startTime > f.dataInfo.startTime))
			dataInfo.startTime = f.dataInfo.startTime;
		if(!dataInfo.endTime.isSet() || (dataInfo.endTime < f.dataInfo.endTime))
			dataInfo.endTime = f.dataInfo.endTime;

		if(userDefinedChannels){
			if((chan + f.dataInfo.channels.size()) > dataInfo.channels.size()){
				return err.set(1, "There are more channels of data in the files than there is station/channel name info");
			}
		}
		
		if(config.sequential && seg && (f.dataInfo.channels.size() != dataInfo.channels.size())){
			return err.set(1, "There are different numbers of channels in the sequential files");
		}

		// Add channels from the file to the list of channels to import
		for(chanFile = 0; chanFile < f.dataInfo.channels.size(); chanFile++){
//			printf("Do Chan: %d,%d %d %d %d\n", chan, seg, chanFile, f.dataInfo.channels.size(), dataInfo.channels.size());
			// If the user has not specifed the channels, add them from the file
			if(!userDefinedChannels){
				if(!config.sequential || !seg){
					dataInfo.channels.append(f.dataInfo.channels[chanFile]);
					if(config.network)
						dataInfo.channels[chan].front().network = config.network;
				}
			}

			importChannels[seg].append(ImportChannel(f, chanFile + 1));

			if(seg == 0){
				dataInfo.channels[chan][0].startTime = f.dataInfo.channels[chanFile].front().startTime;
				dataInfo.channels[chan][0].endTime = f.dataInfo.channels[chanFile].front().endTime;
				dataInfo.channels[chan][0].sampleRate = f.dataInfo.channels[chanFile].front().sampleRate;
				dataInfo.channels[chan][0].sampleFormat = f.dataInfo.channels[chanFile].front().sampleFormat;
				dataInfo.channels[chan][0].importFormat = f.dataInfo.channels[chanFile].front().importFormat;
				dataInfo.channels[chan][0].numSamples = f.dataInfo.channels[chanFile].front().numSamples;
				dataInfo.channels[chan][0].numBlocks = f.dataInfo.channels[chanFile].front().numBlocks;
				dataInfo.channels[chan][0].info = f.dataInfo.channels[chanFile].front().info;
				dataInfo.channels[chan][0].dataFileChannel = (chan + 1);
			}
			else {
				if(config.validateMissingBlocks){
					if(dataInfo.channels[chan][0].endTime != f.dataInfo.channels[chanFile].front().startTime)
						return err.set(1, BString("Data files are not sequential in time: ") + f.dataInfo.channels[chanFile].front().startTime.getString() + " != " + dataInfo.channels[chan][0].endTime.getString());
				}
				else {
					if(dataInfo.channels[chan][0].endTime > f.dataInfo.channels[chanFile].front().startTime)
						return err.set(1, BString("Data files have time overlap: ") + f.dataInfo.channels[chanFile].front().startTime.getString() + " < " + dataInfo.channels[chan][0].endTime.getString());
				}

				dataInfo.channels[chan][0].endTime = f.dataInfo.channels[chanFile].front().endTime;
				dataInfo.channels[chan][0].numSamples += f.dataInfo.channels[chanFile].front().numSamples;
				dataInfo.channels[chan][0].numBlocks += f.dataInfo.channels[chanFile].front().numBlocks;
			}

#if BDS_FUDGE
			// Special case for SEED data where location is set incorrectly
			if((config.format == "SEED") && (dataInfo.startTime >= BTimeStamp("1991-07-10")) && (dataInfo.startTime < BTimeStamp("1995-09-01"))){
				BString	chanType;
				BString	chanAux;

				bdsChannelGetTypeAux(dataInfo.channels[chan].front().channel, chanType, chanAux);
				if(chanAux == "YK"){
					chanAux = "";
					dataInfo.channels[chan][seg].channel = bdsChannelGetName(chanType, chanAux);
					dataInfo.info[BString("importWarning") + warningNumber++] = "BDS_FUDGE, ChannelAux field cleared as it was set to YK";
				}
			}
#endif

			// Set to CM format if sample rate differs per channel
			if(sampleRate == 0.0)
				sampleRate = f.dataInfo.channels[chanFile].front().sampleRate;
			
			if(sampleRate != f.dataInfo.channels[chanFile].front().sampleRate){
				format = "API-CM";
			}

			chan++;
		}

		// Add DataInfo file information
		for(f.dataInfo.info.start(di); !f.dataInfo.info.isEnd(di); f.dataInfo.info.next(di)){
			BString	key;
			
			key = BString("importFile") + fileNumber + "." + f.dataInfo.info.key(di);
			dataInfo.info[key] = f.dataInfo.info[di];
		}
		fileNumber++;
		
		if(config.sequential){
			chan = 0;
			seg++;
		}
	}

	if(importChannels[0].size() != dataInfo.channels.size())
		return err.set(1, "The number of data channels in the files is less than the number of channels configured");

#ifdef ZAP
	bdsDumpDataInfo(dataInfo);
#endif

	// Synchronous
	if((importFiles.size() == 1) && (importFiles[0].dataInfo.synchronous))
		dataInfo.synchronous = 1;
	else
		dataInfo.synchronous = config.synchronous;
	
	// Validate the Metadata if required
	if(err = validateMetaData(bds, dataInfo)){
		errString = BString("MetaData not present, ") + ", " + err.getErrorNo() + ", " + err.getString();
		if(config.validateMetaData){
			printError(errString);
			err.set(ErrorValidate, "Validation error");
			return err;
		}
		else {
			dataInfo.info[BString("importWarning") + warningNumber++] = errString;
			err.clear();
		}
	}

	// If a dry run, stop now
	if(config.dryRun)
		return err;

//	std::cout << "Output: StartTime: " << dataInfo.startTime.getString() <<	" EndTime: " << dataInfo.endTime.getString() << "\n";

	// Setting up data stream
	if(err = bds.dataOpen(dataInfo, "w", format, dataHandle))
		return err;

	// Setup info on this set of files
	if(err = bds.dataSetInfo(dataHandle, dataInfo))
		return err;

	// Read blocks from files
	if(format == "API-SM"){
		// Sample multiplexed import
		if(importFiles.size() != 1)
			return err.set(1, "Only one file can be imported into a sample multiplexed file");
		
		while(!err){
//			printf("Chan: %d %s\n", chan, importFiles.front().dataInfo.channels.front().front().channel.retStr());

			if(err = importFiles.front().file->readData(0, 0, blockNumber, dataBlock))
				break;

//			printf("GotBlock: Chan: %d Block: %d Size: %d\n", chan, blockNumber, dataBlock.channelData[0].size());
			dataBlock.channelNumber = 1;

			if(err = bds.dataPutBlock(dataHandle, dataBlock))
				break;

			blockNumber++;

			if(config.numBlocks && (blockNumber >= config.numBlocks))
				break;
		}

		if(err.getErrorNo() == ErrorEndOfFile)
			err.set(0, "");

	}
	else if(format == "API-CM"){
		// Channel multiplexed import, one channel after another
		for(chan = 0; !err && (chan < importChannels[0].size()); chan++){
			for(seg = 0; !err && (seg < importChannels.size()); seg++){
				blockNumber = 0;

				while(1){
//					printf("WriteBlock: %d,%d %d\n", chan, seg, blockNumber);

					if(err = importChannels[seg][chan].file.file->readData(importChannels[seg][chan].fileChannel, 0, blockNumber, dataBlock)){
						if((err.getErrorNo() == ErrorEndOfData) || (err.getErrorNo() == ErrorEndOfFile))
							err.set(0, "");
						break;
					}

//					printf("GotBlock: Chan: %d NumChans: %d Block: %d Size: %d\n", chan, dataBlock.channelData.size(), blockNumber, dataBlock.channelData[0].size());
					dataBlock.channelNumber = chan + 1;
					dataBlock.segmentNumber = 0;

					if(err = bds.dataPutBlock(dataHandle, dataBlock))
						break;

					blockNumber++;

					if(config.numBlocks && (blockNumber >= config.numBlocks))
						break;
				}
				
				if(err.getErrorNo() == ErrorEndOfFile)
					err.set(0, "");
			}
		}
	}
	else {
		err.set(1, "Unknown import format");
	}

	if(e = bds.dataClose(dataHandle)){
		if(!err)
			err = e;
	}
		
	return err;
}

BError importDataSeedTar(DataAddAccess& bds, Config& config){
	BError		err;
	BString		fileName;
	char		tmpDirPath[] = "/tmp/bdsImportXXXXXX";
	BString		tmpDir;
	int		e;
	BDir		dir;
	BIter		i;
	
	// Perform validation
	if(config.fileNames.size() != 1)
		return err.set(ErrorParam, "Requires one file for SEED-TAR format");

	fileName = config.fileNames.front();
	if(config.verbose)
		printf("Process SEED-TAR File: %s\n", fileName.retStr());

	if(access(fileName, R_OK))
		return err.set(ErrorFile, BString("Unable to access: ") + fileName);

	// Un-tar the files contents
	tmpDir = mkdtemp(tmpDirPath);	
	if(tmpDir == "")
		return err.set(1, BString("Unable to create temporary directory"));
	
	if(e = system(BString("tar -xC ") + tmpDir + " -f " + fileName)){
		system(BString("rm -fr ") + tmpDir);
		return err.set(ErrorFile, BString("Tar un-packing of archive failed: ") + fileName);
	}
	
	if(err = dir.open(tmpDir)){
		system(BString("rm -fr ") + tmpDir);
		return err.set(1, BString("Unable to open temporary directory: ") + tmpDir + " " + err.getString());
	}

	dir.setSort(1);
	if(err = dir.read()){
		system(BString("rm -fr ") + tmpDir);
		return err.set(1, BString("Unable to open temporary directory path: ") + err.getString());
	}

	if(config.sequential){
		// Create a single import file
		config.format = "SEED";
		config.fileNames.clear();
		for(dir.start(i); !dir.isEnd(i); dir.next(i)){
			BString fileName = tmpDir + "/" + dir.entryName(i);

			if(dir.entryName(i)[0] == '.')
				continue;

			config.fileNames.append(fileName);
		}

		err = importData(bds, config);
	}
	else {
		// Process each file in turn
		config.format = "SEED";
		for(dir.start(i); !dir.isEnd(i); dir.next(i)){
			BString fileName = tmpDir + "/" + dir.entryName(i);

			if(dir.entryName(i)[0] == '.')
				continue;

			config.fileNames.clear();
			config.fileNames.append(fileName);

			if(err = importData(bds, config))
				break;
		}
	}

	system(BString("rm -fr ") + tmpDir);
	
	return err;
}

BError formatList(DataAddAccess& bds){
	BError			err;
	BList<DataFormat>	list;
	BIter			i;
	const char*		title[] = { "Formats", "Read/Write", "Description", 0 };
	BTable			table;
	BStringArray		d;
	BString			s;

	table.setTitle(charToArray(title));

	if(err = bds.dataFormatGetList(list))
		return err;

	for(list.start(i); !list.isEnd(i); list.next(i)){
		d.resize(3);
		
		d[0] = blistToString(list[i].names);
		s = "";
		if(list[i].read)
			s += "R";
		if(list[i].write)
			s += "W";
		d[1] = s;
		d[2] = list[i].description;
		table.addRow(d);
	}
	
	d[0] = "SEED-TAR";
	d[1] = "R";
	d[2] = "SEED files in TAR archive";
	table.addRow(d);

	table.print();

	return err;
}


void usage(void) {
	cerr << "Usage:\tbdsImportData [options] <inFileName> <inFileName> ...\n";
	cerr << " -help                - Help on command line parameters\n";
	cerr << " -verbose             - Verbose messages\n";
	cerr << " -host <hostName>     - The BDS server host name\n";
	cerr << " -user <user:password> - user id and password\n";
	cerr << " -network <network> - Network organisation\n";
	cerr << " -array <station>     - The array name, if from an array\n";
	cerr << " -source <source>     - The data source\n";
	cerr << " -format <type>       - The file type: BDRS,WRA\n";
	cerr << " -formatList          - List the supported import formats\n";
	cerr << " -numBlocks <numBlocks> - The number of blocks to write\n";
	cerr << " -channels <station1:chan1,station2:chan2> - The channel information\n";
	cerr << " -sequential          - The data files are sequential in time, import them into a single file\n";
	cerr << " -synchronous         - The channels are synchronously sampled\n";
	cerr << " -ignoreAll           - Allow import although validation fails\n";
	cerr << " -ignoreMissingBlocks - Allow import although validation fails with missing blocks of data\n";
	cerr << " -ignoreTimeBackwards - Allow import although validation fails with blocks going backwards in time\n";
	cerr << " -ignoreFilenameTime  - Allow import although validation fails with file name times that do not match block timestamps\n";
	cerr << " -ignoreMetaData      - Allow import although validation fails where Database metatdata does not match files metadata or is not present\n";
	cerr << " -ignoreCorruptions   - Fix block curruptions (short/long or otherwise corrupted blocks) by ignoring them (BDRS, WRA-40 and WRA-64\n";
	cerr << " -reorder             - Reorder block's in time stamp order. Usefull for SCREAM sourced GCF files where backfilling has occured\n";
	cerr << " -deleteDuplicates    - Delete Duplicate blocks. Usefull for SCREAM sourced GCF files where backfilling has occured\n";
	cerr << " -addWarning <str>    - Add an importWarning string\n";
	cerr << " -warnings            - Display warnings\n";
	cerr << " -dryRun              - Perform all data validation tests but don't actually import the data\n";
	cerr << " -description <str>   - Add an import description\n";
}

static struct option options[] = {
		{ "?",			0, NULL, 0 },
		{ "h",			0, NULL, 0 },
		{ "help",		0, NULL, 0 },
		{ "verbose",		0, NULL, 0 },
		{ "host",		1, NULL, 0 },
		{ "user",		1, NULL, 0 },
		{ "network",		1, NULL, 0 },
		{ "array",		1, NULL, 0 },
		{ "source",		1, NULL, 0 },
		{ "format",		1, NULL, 0 },
		{ "formatList",		0, NULL, 0 },
		{ "channels",		1, NULL, 0 },
		{ "synchronous",	0, NULL, 0 },
		{ "numBlocks",		1, NULL, 0 },
		{ "ignoreAll",		0, NULL, 0 },
		{ "ignoreMissingBlocks",0, NULL, 0 },
		{ "ignoreTimeBackwards",0, NULL, 0 },
		{ "ignoreFilenameTime",	0, NULL, 0 },
		{ "ignoreMetaData",	0, NULL, 0 },
		{ "ignoreCorruptions",	0, NULL, 0 },
		{ "reorder",		0, NULL, 0 },
		{ "deleteDuplicates",	0, NULL, 0 },
		{ "addWarning",		1, NULL, 0 },
		{ "warnings",		0, NULL, 0 },
		{ "dryRun",		0, NULL, 0 },
		{ "description",	1, NULL, 0 },
		{ "sequential",		0, NULL, 0 },
		{ 0,0,0,0 }
};

int main(int argc, char** argv){
	BError			err;
	Config			config;
	int			optIndex = 0;
	int			c;
	BString			s;
	BIter			i;
	DataAddAccess		bds;
	BString			user;
	BString			password;

	config.hostName = getenv("BDS_HOST");
	if(config.hostName == "")
		config.hostName = "localhost";
	
	config.source = "Main";
	config.numBlocks = 0;
	config.synchronous = 0;
	config.verbose = 0;
	config.validate = 1;
	config.validateMissingBlocks = 1;
	config.validateTimeBackwards = 1;
	config.validateFilenameTime = 1;
	config.validateMetaData = 1;
	config.validateCorruptions = 1;
	config.reorder = 0;
	config.deleteDuplicates = 0;
	config.warnings = 0;
	config.dryRun = 0;
	config.formatList = 0;
	config.sequential = 0;
	
	// Disable DataFileSeed's locking, we will manage this
	DataFileSeed::onoLock = 1;

	while((c = getopt_long_only(argc, argv, "", options, &optIndex)) == 0){
		s = options[optIndex].name;
		if(s == "help" || s == "h" || s == "?"){
			usage();
			return 1;
		}
		else if(s == "verbose"){
			config.verbose = 1;
		}
		else if(s == "host"){
			config.hostName = optarg;
		}
		else if(s == "user"){
			s = optarg;
			user = s.pullToken(":");
			password = s;
		}
		else if(s == "network"){
			config.network = optarg;
		}
		else if(s == "array"){
			config.array = optarg;
		}
		else if(s == "source"){
			config.source = optarg;
		}
		else if(s == "formatList"){
			config.formatList = 1;
		}
		else if(s == "format"){
			config.format = optarg;
		}
		else if(s == "channels"){
			config.channels = bstringToList(optarg);
		}
		else if(s == "synchronous"){
			config.synchronous = 1;
		}
		else if(s == "sequential"){
			config.sequential = 1;
		}
		else if(s == "numBlocks"){
			config.numBlocks = atoi(optarg);
		}
		else if(s == "ignoreAll"){
			config.validate = 0;
		}
		else if(s == "ignoreMissingBlocks"){
			config.validateMissingBlocks = 0;
		}
		else if(s == "ignoreTimeBackwards"){
			config.validateTimeBackwards = 0;
		}
		else if(s == "ignoreFilenameTime"){
			config.validateFilenameTime = 0;
		}
		else if(s == "ignoreMetaData"){
			config.validateMetaData = 0;
		}
		else if(s == "ignoreCorruptions"){
			config.validateCorruptions = 0;
		}
		else if(s == "reorder"){
			config.reorder = 1;
		}
		else if(s == "deleteDuplicates"){
			config.deleteDuplicates = 1;
		}
		else if(s == "addWarning"){
			config.warningList.append(optarg);
		}
		else if(s == "warnings"){
			config.warnings = 1;
		}
		else if(s == "description"){
			config.description = optarg;
		}
		else if(s == "dryRun"){
			config.dryRun = 1;
		}
		else {
			usage();
			return 1;
		}
	}

	if(c == '?'){
		usage();
		return 1;
	}
	
	while(argc - optind){
		config.fileNames.append(argv[optind++]);
	}


	// Open BDS access
	if(!err){
		err = bds.connectService(BString("//") + config.hostName + "/bdsDataAddAccess");
	}

	// Connect to service
	if(!err)
		err = bds.connect(user, password);

	if(!err){
		if(config.formatList){
			err = formatList(bds);
		}
		else {
			if(config.format == ""){
				err.set(1, "Requires format to be set");
			}

			if(config.fileNames.number() < 1)
				err.set(1, "Requires at least one file to work with");

			if(config.format == "SEED-TAR"){
				err = importDataSeedTar(bds, config);
			}
			else {
				err = importData(bds, config);
			}
		}
	}
	
	if(err)
		printError(err.getString());

	if(err.getErrorNo() < 0)
		return 1;
	else
		return err.getErrorNo();
}
